# 2024.7.1 version.3
# Author: Jin Qian
# This documentation is for converting models. 
# Based on mobilenetv2 example
# From PyTorch to Onnx, based on images as well.
# torchvision: '0.8.2+cpu'

#https://doc.kneron.com/docs/#plus_python/api_document/V2.3.0/

# 首先使用下面的命令打开data1文件夹
# docker run --rm -it -v ~/data1:/data1 kneron/toolchain:latest

# 现在三个值之间的difference很大，但是预测的probability可能是对的

import torch
import torch.onnx
import onnx
import ktc
from PIL import Image
import numpy as np
import os

import torch.nn as nn
import torchvision.models as models
import torchvision.transforms as transforms

from torchvision import models
import onnxruntime as ort
import cv2

def softmax(x, axis=1):
    e_x = np.exp(x)
    return e_x / e_x.sum(axis=axis, keepdims=True)

def preprocess(input_file):
    image = Image.open(input_file)
    image = image.convert("RGB")
    img_data = np.array(image.resize((224, 224), Image.BILINEAR)) / 255
    img_data = np.transpose(img_data, (2, 0, 1))
    img_data = np.expand_dims(img_data, 0)
    return img_data

def inference_comparison(a_probs, b_probs, a_labels, b_labels):
    prob_diff = np.abs(a_probs - b_probs).max()
    label_diff = np.abs(a_labels - b_labels).max() 
    return prob_diff, label_diff

# TODO: 写一个保存结果的函数，后面在结果保存的时候调用，这样不用每次都重写代码
def save_results(labels, probs, model_name):
    with open('/data1/kl720_example_convert_code+data/example_convert/results_example_mobilenetV2.txt', 'a') as f:
        f.write(f"{model_name} Labels: {labels}\n")
        f.write(f"{model_name} Probabilities: {probs}\n\n")

def save_differences(labels, probs, model_name):
    with open('/data1/kl720_example_convert_code+data/example_convert/results_example_mobilenetV2.txt', 'a') as f:
        f.write(f"Label difference{model_name} : {labels}\n")
        f.write(f"Probability difference{model_name} : {probs}\n\n")

raw_images = os.listdir("/data1/example_images/")
input_images = [preprocess("/data1/example_images/" + image_name) for image_name in raw_images]

######## 1. pytorch模型 #########
model = models.mobilenet_v2()
state_dict = torch.load("/data1/kl720_example_convert_code+data/example_convert/save_models/mobilenet_v2-b0353104.pth", map_location='cpu')
model.load_state_dict(state_dict)
model.eval()  # 设置模型为评估模式

# 1. 得到pytorch预测结果
torch_predictions = []
input_tensors = [torch.from_numpy(image).float() for image in input_images]
for input_data in input_tensors:
    with torch.no_grad():
        pred = model(input_data)
    torch_predictions.append(pred.numpy())
torch_predictions = np.vstack(torch_predictions)
torch_predictions = softmax(torch_predictions)
torch_labels = torch_predictions.argmax(axis=1)
torch_probs = torch_predictions.max(axis=1)
##TODO：保存labels和probs
with open('/data1/kl720_example_convert_code+data/example_convert/results_example_mobilenetV2.txt', 'w') as f:
    f.write(f"Pytorch Label: {torch_labels}\n")
    f.write(f"Pytorch probability: {torch_probs}\n\n")

######### 2. 由pytorch转pytorch模型 #########
# 保存onnx模型 到/data1/example_convert/save_models/mobilenetv2_example.onnx
dummy_input = torch.randn(1, 3, 224, 224)
torch.onnx.export(model, dummy_input, '/data1/kl720_example_convert_code+data/example_convert/save_models/mobilenetv2_example.onnx', opset_version=11)

# 2. 得到由pytorch转化的onnx的预测结果
# 使用 ONNX Runtime 进行推理
onnx_model_path = '/data1/kl720_example_convert_code+data/example_convert/save_models/mobilenetv2_example.onnx'
onnx_session = ort.InferenceSession(onnx_model_path)

# 获取 ONNX 模型的输入名
input_name = onnx_session.get_inputs()[0].name

onnx_predictions = []
for input_data in input_images:
    input_data = np.expand_dims(input_data, axis=0).astype(np.float32)  # 增加 batch 维度并转换为 float32
    input_data = input_data.squeeze(axis=0) 
    pred = onnx_session.run(None, {input_name: input_data})[0]
    onnx_predictions.append(pred)
    
onnx_from_torch_predictions = np.vstack(onnx_predictions).reshape(10,1000)
onnx_predictions = softmax(onnx_from_torch_predictions)
#onnx_predictions = torch.softmax(torch.tensor(onnx_from_torch_predictions), dim=1).numpy()
onnx_labels = onnx_predictions.argmax(axis=1)
onnx_probs = onnx_predictions.max(axis=1)
##TODO：保存labels和probs
save_results(onnx_labels, onnx_probs, "Onnx")

######### 结果比较A： pytorch转onnx模型的正确性 #########
prob12_diff, label12_diff = inference_comparison(torch_predictions, onnx_predictions, torch_labels, onnx_labels)
#TODO: 将结果保存下来
save_differences(label12_diff, prob12_diff, "(pytorch to onnx)")

######### 3. optimized onnx模型 #########
exported_m = onnx.load('/data1/kl720_example_convert_code+data/example_convert/save_models/mobilenetv2_example.onnx')
result_m = ktc.onnx_optimizer.torch_exported_onnx_flow(exported_m)
optimized_m = ktc.onnx_optimizer.onnx2onnx_flow(result_m, eliminate_tail=True, opt_matmul=False)
onnx.save(optimized_m, '/data1/kl720_example_convert_code+data/example_convert/save_models/optimized.onnx')

# 3. optimized onnx模型的预测结果
optonnx_predictions = []
for input_data in input_images:
    pred = ktc.kneron_inference(input_data, onnx_file="/data1/kl720_example_convert_code+data/example_convert/save_models/optimized.onnx", input_names=["input.1"])   
    optonnx_predictions += pred
optonnx_predictions = np.vstack(optonnx_predictions)
optonnx_predictions = softmax(optonnx_predictions)
optonnx_labels = optonnx_predictions.argmax(axis=1)
optonnx_probs = optonnx_predictions.max(axis=1)
##TODO：保存labels和probs
save_results(optonnx_labels, optonnx_probs, "Optimized onnx")

######### 结果比较B： onnx转optonnx模型的正确性 #########
prob23_diff, label23_diff = inference_comparison(onnx_predictions, optonnx_predictions, onnx_labels, optonnx_labels)
#TODO: 将结果保存下来
save_differences(label23_diff, prob23_diff, "(onnx to optimized)")

######### 4. bie模型 #########
# IP Evaluation
km = ktc.ModelConfig(32770, "0004", "720", onnx_model=optimized_m)
eval_result = km.evaluate(output_dir="/example_convert")
# 会在终端生成一个reprot
# /data1/kneron_flow/model_fx_report.json

# Quantization
input_mapping = {"input.1": input_images}
bie_path = km.analysis(input_mapping, threads = 4)
# Save .bie file in /data1/kneron_flow/input.kdp720.scaled.bie
# 系统会自己生成一个kneron_flow的文件夹

# 4. bie模型的预测结果
bie_predictions = []
for input_data in input_images:
    pred = ktc.kneron_inference(input_data, bie_file=bie_path, input_names=["input.1"], platform = "720")
    bie_predictions += pred
bie_predictions = np.vstack(bie_predictions)
bie_predictions = softmax(bie_predictions)
bie_labels = bie_predictions.argmax(axis=1)
bie_probs = bie_predictions.max(axis=1)
##TODO：保存labels和probs
save_results(bie_labels, bie_probs, "Bie")

######### 结果比较C： optonnx转bie模型的正确性 #########
prob34_diff, label34_diff = inference_comparison(optonnx_predictions, bie_predictions, optonnx_labels, bie_labels)
#TODO: 将结果保存下来
save_differences(label34_diff, prob34_diff, "(optimized to bie)")

######### 5. nef模型 #########
# Compilation
compile_result = ktc.compile([km])
# Save .net file in /data1/kneron_flow/models_720.nef

# 5. nef模型的预测结果
nef_predictions = []
for input_data in input_images:
    pred = ktc.kneron_inference([input_data], nef_file=compile_result, input_names=["input.1"], model_id=32770, platform=720)
    nef_predictions.append(pred)
nef_predictions = np.vstack(nef_predictions).reshape(len(input_images), -1)
nef_predictions = softmax(nef_predictions)
nef_labels = nef_predictions.argmax(axis=1)
nef_probs = nef_predictions.max(axis=1)
##TODO：保存labels和probs
save_results(nef_labels, nef_probs, "Nef")

######### 结果比较D： bie转nef模型的正确性 #########
prob45_diff, label45_diff = inference_comparison(bie_predictions, nef_predictions, bie_labels, nef_labels)
#TODO: 将结果保存下来
save_differences(label45_diff, prob45_diff, "(bie to nef)")

######### 结果比较E： torch与nef模型的比较 #########
prob15_diff, label15_diff = inference_comparison(torch_predictions, nef_predictions, torch_labels, nef_labels)
#TODO: 将结果保存下来
save_differences(label15_diff, prob15_diff, "(torch to nef)")
